******************************************************
*  Program...........: benford.prg
*  Author............: Lauren Clarke (lauren@cornerstonenw.com)
*  Project...........: csAudit
*  Created...........: 01/10/2002
*  Copyright.........: (c) Cornerstone Consulting, 1999-2000. All rights reserved.
*  Description.......: Demo script for FT Benford article
*					    
*  Unit Test.........: None
*  Dependencies......: stats.prg; xmlgrapher.prg; tastrade data
*
*  Last Change.......: 
******************************************************
* THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
* INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND
* FITNESS FOR A PARTICULAR PURPOSE. THE AUTHOR SHALL NOT BE LIABLE FOR ANY DAMAGE 
* CAUSED BY ITS USE, INCLUDING BUT NOT LIMITED TO LOSS OF PROFIT, LOSS OF DATA, 
* AND LOSS OF VALUE. IN THE EVENT THIS PACKAGE IS FOUND TO BE DEFECTIVE, 
* THE AUTHOR SHALL NOT BE LIABLE FOR ANY DAMAGE CAUSED BY ITS USE, INCLUDING BUT 
* NOT LIMITED TO LOSS OF PROFIT, LOSS OF DATA, 
* AND LOSS OF VALUE EVEN AFTER THE AUTHOR IS MADE AWARE OF SUCH DEFECTS.


******************************************************
* General Comments
*
* You will need MSIE 5.0 or better to render the VML plots generated below.
*
* Run this prg and your browser will open with three result plots:
* FirstDigit, SecondDigit, and First2Digit distributions.
*
* This prg is a scripted example of some benford tests and charting
* If you trace through this script once it should give a good idea 
* of what is going on. 
* 
*
******************************************************


set procedure to stats.prg
set procedure to xmlgrapher.prg additive

local lcTableName, lcFieldName, loGraph, loTTOrders, ;
      loTTFDist, loBenfordStat, loBenfordFD, lcStr

* set these to any .dbf/field you want
lcTableName = "orders"
lcFieldName = "order_net"

*-- Benford Calculation Object
loBenfordStat = createobject( "BenfordCalc" )

*-- Extract Pre-defined Benford distribution
select nvalue,np ;
  from benford ;
  where cposition=="FIRST" ;
  order by nvalue ;
  into array laTemp

*-- Establish First Digit expected distribution
loBenfordFD = createobject( "DataSet" )
loBenfordFD.setData( @laTemp,1,2)

*-- Set graphing options
loBenfordFD.cType  = [XY]
loBenfordFD.nLineWidth=0.5

*-- Set up a Data set to analize
loTTOrders    = createobject( "DataSet" )

if !used(lcTableName)
  use _samples+"data\"+lcTableName+".dbf" in 0
endif

loTTOrders.setData( lcTableName, lcFieldName)

*-- Calculate the first digit distribution

loTTFDdist = loBenfordStat.digitDist( loTTOrders, "FIRST")



*-- Set graphing options
with loTTFDdist
  .cType=[BAR]
  .nLineWidth=10
  .cLineColor="#4682B4"
endwith


*-- calcuate the z-stat boundaries (currently defaulted to 5%)
loZstatUp = Zstat(loBenfordFD ,  loTTOrders.totalPoints(), [UPPER])
loZstatLo = Zstat(loBenfordFD ,  loTTOrders.totalPoints(), [LOWER])

with loZstatUp
  .cType=[XY]
  .nLineWidth=0.5
  .cPointStyle=[]
  .cLineColor = "Darkgoldenrod"
  .cDashStyle = "dashdot"
endwith

with loZstatLo
  .cType=[XY]
  .nLineWidth=0.5
  .cPointStyle=[]
  .cLineColor = "Darkgoldenrod"
  .cDashStyle = "dashdot"
endwith


*-- Graph generator with VML rendering strategy
loGraph = createobject("SMGrapher","VMLArtist")

with loGraph
  .nMaxXcoord = 9.3
  .nMaxYcoord = 0.4
  .nHeight    = 150
  .nWidth     = 300
  .cBGColor   = "cornsilk"
  .nYlabelScale = 0.1
  .nYScale    = 0.1
  .nXScale    = 1
  .nXlabelScale = 1

  .cTitle     = [First Digit Distribution: ]+lcTableName+[.]+lcFieldName
  
  .addDataSet( loTTFDdist )
  .addDataSet( loBenfordFD )
  .addDataSet( loZStatUp )
  .addDataSet( loZStatLo )
endwith

*-- At this point loGraph is ready to go with the first digit analysis

*--- 2nd digit test

*-- determine second digit distribution of sample data
loTTSDdist = loBenfordStat.digitDist( loTTOrders, "SECOND")

*-- set graphing options
with loTTSDdist
  .cType=[BAR]
  .nLineWidth=10
  .cLineColor="#4682B4"
endwith

*-- extract pre-determined Benford distribution for second digit
select nvalue,np ;
  from benford ;
  where cposition=="SECOND" ;
  order by nvalue ;
  into array laTemp

loBenfordSD = createobject( "DataSet" )

with loBenfordSD
  .setData( @laTemp,1,2)
  .cType  = [XY]
  .nLineWidth=0.5
endwith

loZstatUp = Zstat(loBenfordSD ,  loTTOrders.totalPoints(), [UPPER])
loZstatLo = Zstat(loBenfordSD ,  loTTOrders.totalPoints(), [LOWER])


with loZstatUp
  .cType=[XY]
  .nLineWidth=0.5
  .cPointStyle=[]
  .cLineColor = "Darkgoldenrod"
  .cDashStyle = "dashdot"
endwith

with loZstatLo
  .cType=[XY]
  .nLineWidth=0.5
  .cPointStyle=[]
  .cLineColor = "Darkgoldenrod"
  .cDashStyle = "dashdot"
endwith


loGraph2 = createobject("SMGrapher","VMLArtist")

with loGraph2
  .nMaxXcoord = 9
  .nMaxYcoord = 0.15
  .nHeight    = 150
  .nWidth     = 300
  .cBGColor   = "cornsilk"
  .nYlabelScale = 0.01
  .nYScale    = 0.01
  .nXScale    = 1
  .nXlabelScale = 1
  .cLables    = [XY]
  .cGridLines = [XY]
  .cTitle     = [Second Digit Distribution: ]+lcTableName+[.]+lcFieldName
  .addDataSet( loTTSDdist)
  .addDataSet( loBenfordSD)
 * .addDataSet( loZStatUp)
 * .addDataSet( loZStatLo )
endwith

*--- First2 digit test

*-- calculate the first2 digit distribution of the sample data
loTTF2dist = loBenfordStat.first2DigitDist( loTTOrders)

with loTTF2dist
  .cType=[BAR]
  .nLineWidth=3
  .cLineColor="#4682B4"
endwith

select nvalue,np ;
  from benford ;
  where cposition=="FIRST2" ;
  order by nvalue ;
  into array laTemp

loBenfordF2 = createobject( "DataSet" )

with loBenfordF2
  .setData( @laTemp,1,2)
  .cType  = [XY]
  .nLineWidth=0.5
  .cPointStyle= []
endwith


loZstatUp = Zstat(loBenfordF2 , loTTOrders.totalPoints(), [UPPER])
loZstatLo = Zstat(loBenfordF2 , loTTOrders.totalPoints(), [LOWER])


with loZstatUp
  .cType=[XY]
  .nLineWidth=0.5
  .cPointStyle=[]
  .cLineColor = "blue"
  .cDashStyle = "dashdot"
endwith

with loZstatLo
  .cType=[XY]
  .nLineWidth=0.5
  .cPointStyle=[]
  .cLineColor = "blue"
  .cDashStyle = "dashdot"
endwith


loGraph3 = createobject("SMGrapher","VMLArtist")

with loGraph3
  .nMaxXcoord = 100
  .nMaxYcoord = 0.065
  .nHeight    = 200
  .nWidth     = 400
  .cBGColor   = "cornsilk"
  .nYlabelScale = 0.01
  .nYScale    = 0.01
  .nXScale    = 10
  .nXlabelScale = 10
  .cLables    = [XY]
  .cGridLines = [XY]
  .cTitle     = [First 2 Digit Distribution: ]+lcTableName+[.]+lcFieldName
  .addDataSet( loTTF2dist)
  .addDataSet( loBenfordF2)
  .addDataSet( loZStatUp)
  .addDataSet( loZStatLo)
endwith


*-- Final Page generation

local lcStr 
lcStr = []

set textmerge on

text to lcStr noshow

<html xmlns:v="urn:schemas-microsoft-com:vml">
<head>
<style>
v\:* { behavior: url(#default#VML); }
.xlabel 
{font-family: "Courier New", Courier, monospace;
font-weight: normal;
font-size: 8pt;
color: black;}

.ylabel 
{font-family: "Courier New", Courier, monospace;
font-weight: normal;
font-size: 8pt;
color: black;}

.charttitle
{font-family: Arial,"Courier New", Courier, monospace;
font-weight: bold;
font-size: 12pt;
color: cadetblue;}

</style>
</head>
<body>
<<loGraph.drawChart()>><br><br>
<<loGraph2.drawChart()>><br><br>
<<loGraph3.drawChart()>>
</body>
</html>

endtext

*-- write the report file
strtofile( lcSTR, "benford.htm")

lcFile = fullpath( "benford.htm" )

DECLARE INTEGER ShellExecute IN "Shell32.dll" ;
  INTEGER hwnd, ;
  STRING lpVerb, ;
  STRING lpFile, ;
  STRING lpParameters, ;
  STRING lpDirectory, ;
  LONG nShowCmd

*-- view the file
=ShellExecute( 0,"Open",lcFile,"","",1 )



