******************************************************
*  Program...........: stats.prg
*  Author............: Lauren Clarke (lauren@cornerstonenw.com)
*  Project...........: csStats
*  Created...........: 11/14/2001
*  Copyright.........: (c) Cornerstone Consulting, 1999-2002. All rights reserved.
*  Description.......: General statistics helper UDF's and Classes
*					    
*  Unit Test.........: stats_ut.prg
*  Dependencies......: 
*
*  Last Change.......: 02/25/2002 (see change log at end of file)
******************************************************
* THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
* INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND
* FITNESS FOR A PARTICULAR PURPOSE. THE AUTHOR SHALL NOT BE LIABLE FOR ANY DAMAGE 
* CAUSED BY ITS USE, INCLUDING BUT NOT LIMITED TO LOSS OF PROFIT, LOSS OF DATA, 
* AND LOSS OF VALUE. IN THE EVENT THIS PACKAGE IS FOUND TO BE DEFECTIVE, 
* THE AUTHOR SHALL NOT BE LIABLE FOR ANY DAMAGE CAUSED BY ITS USE, INCLUDING BUT 
* NOT LIMITED TO LOSS OF PROFIT, LOSS OF DATA, 
* AND LOSS OF VALUE EVEN AFTER THE AUTHOR IS MADE AWARE OF SUCH DEFECTS.


******************************************************
* General Comments
*
******************************************************


*---------------------------------------------
function Zstat( toSet,tn,tcbound )
* returns a reference to a set with a zstat for each stat in toSet
*? works only with 2nd col

if vartype( tcBound ) # "C"
  tcBound="UPPER"
endif

tcBound = alltrim(upper( tcBound ) )

 
local lnTotal, lnBound

lnBound = iif( tcBound="LOWER",-1,1)

lnTotal = toSet.totalPoints()


if lnTotal =0
  error "Zstat function failed: no data"
endif

local loResult,laResult(lnTotal,2), lnK,lnEx

loResult = createobject( "DataSet" )
store 0 to laResult

for lnK = 1 to lnTotal
  lnEx = toset.getCoord(lnK,2)
  laResult(lnk,1) = toset.getCoord(lnK,1)
  laResult(lnK,2) = lnEx +lnBound * 1.96 * sqrt( lnEx * (1-lnEx)/tn)+lnBound/(2*tn)
 
endfor

loresult.setData( @laResult,1,2)

return loresult



*---------------------------------------------
function MAD( toSet1, toSet2, taErrors )
* calculates absolute deviations between toSet1 and toSet2
* returns the Mean absolute 
* returns -1 for error
*? assumes 2nd column (y) data

local lnTotal

lnTotal = toSet1.totalPoints()
if lnTotal <> toSet2.TotalPoints() or lnTotal = 0
  error "MAD function failed: no data or data mismatch"
endif

local lnK, lnSum, lnAbsError

dimension taErrors( lnTotal )

lnSum = 0

for lnK = 1 to lnTotal
  lnVal1  =   toSet1.getCoord(lnK,2)
  lnVal2  =   toSet2.getCoord(lnK,2) 
  
  lnAbsError = abs( lnVal1 - lnVal2 )
  lnSum      = lnSum + lnAbsError
  taErrors(lnK) = lnAbsError
endfor

return lnSum / lnTotal




*=============================================
define class BenfordCalc as Relation
*=============================================
* performs various DA analysis based on Benford's law

*---------------------------------------------
function Digit( tvNumber, tnSig , tnLen )
local lcNum, lnNum

if vartype( tnLen ) # "N"
  tnLen = 1
endif

*-- standardize vartype
lnNum = xVal( tvNumber )

*-- convert currency
if vartype( lnNum ) ="Y"
  lnNum = mton(lnNum)
endif

*-- standardize mantissa
if !empty(lnNum)
  lnNum = abs(lnNum / ;
          10^(round(log10(abs(lnNum))-1,0)))
endif

lcNum = strtran(transform(lnNum),".")

*-- strip leading zeros
*!*	do while !empty(lcNum) and left(lcNum,1)="0"
*!*	  lcNum = substr(lcNum,2)
*!*	enddo
return val( substr( lcNum ,tnSig,tnLen) )

*---------------------------------------------
function First2DigitDist( toData )
local lnTotal, loResult, laResult(90,2), lnK, lotemp

loResult = createobject( "DataSet" )
store 0 to laResult

*-- we include 0 for standardization with other measures
for lnK = 1 to 90
  laResult(lnK,1) = lnK+9
endfor

lnTotal = toData.TotalPoints()

local lnSkip
lnSkip = 0

if lnTotal > 0 

  for lnK = 1 to lnTotal
  
    lnTemp    = toData.getCoord(lnK,1)

    lnDigit = this.Digit( lnTemp, 1 , 2 )     
    
    if lnDigit = 0
      lnSkip = lnSkip + 1 
      loop
    endif
    
    lnRow   = lnDigit - 9
    
    
    laResult( lnRow,2 ) = laResult( lnRow,2) + 1
    
  endfor
  
  for lnK = 1 to 90
    laResult(lnK,2) = laResult(lnK,2)/(lnTotal-lnSkip)
  endfor
  
endif

loResult.setData( @laResult, 1, 2)

return loResult
  
  
*---------------------------------------------        
function DigitDist( toData, lcDigit)
* returns object dataset with first digit distribution
*? fix for multi-column
local lnTotal, loResult, laResult(10,2), lnK, lotemp

loResult = createobject( "DataSet" )
store 0 to laResult
lcDigit = upper(alltrim( lcDigit ))

*-- we include 0 for standardization with other measures
for lnK = 1 to 10
  laResult(lnK,1) = lnK-1
endfor

lnTotal = toData.TotalPoints()

if lnTotal > 0 

  for lnK = 1 to lnTotal

    lnTemp    = toData.getCoord(lnK,1)
    
    do case
      case lcDigit $ "FIRST,1" 
        lnDigit = this.Digit( lnTemp, 1 )
      case lcDigit $ "SECOND,2"
        lnDigit = this.Digit( lnTemp, 2 )

      otherwise
        error "Unrecognized Digit type in BenfordCalc::DigitDist"
    endcase
    
    lnRow   = lnDigit + 1
    laResult( lnRow,2 ) = laResult( lnRow,2) + 1
    

  endfor

  for lnK = 1 to 10
    laResult(lnK,2) = laResult(lnK,2)/lnTotal
  endfor
endif

loResult.setData( @laResult, 1, 2)

return loResult


enddefine





*=============================================
define class AbsGammaDist as Relation
*=============================================

cTableName    = ""
nError        = 0
cErrorMessage = ""

*---------------------------------------------
function opentable()

with this
  if !used( .cTableName )
    use (.cTableName ) in 0
  else
    select ( .cTableName )   
  endif
endwith


function calcStat( toData, tnDataColumn, toExpected, tnExpColumn)

function test( toData, toExpected, tnPval )

function tablelookup( tnDF, tnPval )


enddefine

*=============================================
define class ChiSquaredDist as AbsGammaDist
*=============================================

cTableName = "chisquared"
nStat      = -1
nCutoff   = -1

*---------------------------------------------
function test( toData, toExpected, tnPval )
local lnCutoff, lnStat

lnStat = this.calcStat( toData, toExpected )

if this.nError>0
  return .f.
endif

lnCutoff = this.tableLookup( tnPval, toData.totalPoints() - 1)

this.nStat = lnStat
this.nCutoff = lnCutoff

return lnStat<lnCutoff

*---------------------------------------------
function calcStat( toData, toExpected)
local lnTotal

lnTotal = toData.totalPoints()
if lnTotal <> toExpected.TotalPoints()
  this.nError = 3
  this.cErrorMessage = "Data Sets do not have equal number of points"
  return -1
endif

local lnSum 

lnSum = 0

*-- calc the chi-squared statistic
for lnK = 1 to lnTotal
  lnActual    = toData.getCoord(lnK,2)
  lnExpected  = toExpected.getCoord(lnK,2)  
  *? needs zero check
  lnSum = lnSum + ;
          ( (lnActual - lnExpected) ^ 2 ) / lnExpected
endfor

return lnSum 
     
*---------------------------------------------
function tablelookup( tnPval, tnDF )

this.openTable()
 
if tnDF> reccount( (this.cTableName) )

  this.nError = 1
  this.cErrorMessage = "Table has insufficient records to support "+alltrim(str(tnDf))+" degrees of freedom"
  return -1

endif

local laFields(1), lcField

*? opt this
afields( laFields, this.cTableName )
lcField = alltrim(transform(tnPval*1000,[###]))
lcField = [P]+iif( len(lcField) = 2, [0],[])+lcField

if ! ascan(laFields,lcField,1,-1,1,1)>0

  this.nError = 2
  this.cErrorMessage = "Requested P value: "+lcField+" is not available in "+this.ctablename
  return -1

endif

go tnDF

return evaluate( alltrim( this.cTableName ) + [.] + lcField )

enddefine


*=============================================
define class dataSet as Relation
*=============================================
*? convert this to proper collection via _access/_assign
*    and allow for add/delete methods etc.
*? support more than 2 cols
*? consider cType strategy...

cDataType 	= [ARRAY]
cTableName	= []
cType       = [XY]
dimension aCols(1)

cLineColor  = [black]
cPointColor = [blue]
cPointStyle = [CIRCLE]
cDashStyle  = []

nLineWidth  = 1
nColCount   = 2

dimension aPoints(1)

*---------------------------------------------
function setData( tvData, tvCol1, tvCol2 )
local lnCols
*? generalize for multiple columns

this.nColCount = pcount() - 1
this.nColCount = iif(this.nColCount=0,1,this.nColCount)

do case 

case vartype( tvData ) = "C"
  this.cTableName = tvData
  this.cDataType=[TABLE]
  local laFields(1)
  
  *-- map the field names to column numbers
  
  if vartype( tvCol1 ) # "N"
    afields( laFields, this.cTableName)
    lnCol1 = ascan(laFields,tvCol1,1,-1,1,9)
  
    if lnCol1 = 0
      error "Field: "+tvCol1+" is not found in: "+this.cTableName
    endif
  else
    lnCol1 = tvCol1
  endif
  
  if this.nColCount>1

  if vartype( tvCol2 ) # "N"
      lnCol2 = ascan(laFields,tvCol2,1,-1,1,9)
  
      if lnCol2 = 0
        error "Field: "+tvCol2+" is not found in: "+this.cTableName
      endif
    else
      lnCol2 = tvCol2
    endif
  
  endif
  
otherwise && assume array
  acopy( tvData, this.aPoints )
  this.cDataType=[ARRAY]
  lnCol1 = tvCol1
  lnCol2 = tvCol2
endcase

dimension this.aCols( this.nColCount )
this.aCols(1) = lnCol1
if this.nColCount>1
this.aCols(2) = lnCol2
endif


*---------------------------------------------
function totalPoints()

do case
case this.cDataType=[ARRAY]
  return iif( vartype( this.aPoints(1) ) = "L", 0 ,alen( this.aPoints,1) )
case this.cDataType=[TABLE]
  return reccount( this.cTableName )
otherwise
  error [unrecognized dataSet type]
endcase



*---------------------------------------------
function getCoord( lnRow, lnCol )
local loPoint,lvRetVal

do case 
  case this.cDataType=[ARRAY]

    if alen(this.aPoints,2)<1
      lvRetval = this.aPoints( lnRow )
    else
      lvRetval = this.aPoints( lnRow, this.aCols(lnCol) )
    endif
    
 case this.cDataType=[TABLE]
   
   select (this.cTableName)

   *-- careful here, we'll find deleted records 
   *-- 02/25/2002 thanks to BP for the fix.
   *-- maybe better to SELECT the collection into a cursor first.
   
   set near on
   go lnRow
   set near off
   If deleted()
     Skip
   endif
   If EOF()
     SKIP -1
   Endif
   
   lvRetval = evaluate( field( this.aCols(lnCol) ) )
    
otherwise
  error [unrecognized dataSet type]
endcase

return lvRetval


*---------------------------------------------
function getPoint( lnRow )
local loPoint

do case 
  case this.cDataType=[ARRAY]
    loPoint = createobject("xyPoint")
    loPoint.x = this.aPoints(lnRow, this.aCols(1) )
    if this.nColCount>1
      loPoint.y = this.aPoints(lnRow, this.aCols(2) )
    endif
    
    
 case this.cDataType=[TABLE]
    
    select (this.cTableName)
    *? trap error
    go lnRow
    
    
    loPoint = createobject("xyPoint")
    loPoint.x = evaluate( field( this.aCols(1) ) )
    
    if this.nColCount>1
      loPoint.y = evaluate( field( this.aCols(2) ) )
    endif
    
otherwise
  error [unrecognized dataSet type]
endcase

return  loPoint

enddefine

*=============================================
define class xyPoint as Relation
*=============================================
*? consider n dimensional via array
x = 0
y = 0
enddefine 

* 02/25/2002 modified dataSet::getCoord() to respect filters and set delete