******************************************************
*  Program...........: xmlGrapher.prg
*  Author............: Lauren Clarke (lauren@cornerstonenw.com)
*  Project...........: csStats
*  Created...........: 11/14/2001
*  Copyright.........: (c) Cornerstone Consulting, 1999-2003. All rights reserved.
*  Description.......: xml graph generator
*					    
*  Unit Test.........: xmlgrapher_ut.prg
*  Dependencies......: stats.prg (needs DataSet)
*
*  Last Change.......: 
******************************************************
* THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
* INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND
* FITNESS FOR A PARTICULAR PURPOSE. THE AUTHOR SHALL NOT BE LIABLE FOR ANY DAMAGE 
* CAUSED BY ITS USE, INCLUDING BUT NOT LIMITED TO LOSS OF PROFIT, LOSS OF DATA, 
* AND LOSS OF VALUE. IN THE EVENT THIS PACKAGE IS FOUND TO BE DEFECTIVE, 
* THE AUTHOR SHALL NOT BE LIABLE FOR ANY DAMAGE CAUSED BY ITS USE, INCLUDING BUT 
* NOT LIMITED TO LOSS OF PROFIT, LOSS OF DATA, 
* AND LOSS OF VALUE EVEN AFTER THE AUTHOR IS MADE AWARE OF SUCH DEFECTS.


******************************************************
* General Comments
*
* SMGrapher is the primary worker here. It utilizes
* an Artist strategy to laydown XML according to the spec
* required. There is only VMLArtist available now, but
* SVG artist would be an obvious second strategy. 
* 
******************************************************


*=============================================
define class AbsArtist as Relation
*=============================================
* abstract strategy, modifies SMGrapher
nMaxX = 100
nMaxY = 100

nWidth  = 100
nHeight = 100

nXmesh = 1000
nYmesh = 1000

nXAR   = 1
nYAR   = 1

*-- pixels per unit
nXPPU  = 1	
nYPPU  = 1


function nXPPU_access
return this.nWidth/this.nMaxX

function nYPPU_access
return this.nHeight/this.nMaxY

function transformX( tx )
return tx

function transformY( ty )
return ty

function coOrdToString(tx,ty)
return( alltrim(str( tx,10,10))+[, ]+alltrim(str( ty,10,10 )) )


function drawLine( tnWidth, tcColor, tnX1, tnY1, tnX2, tnY2 )

function drawPoint( tnWidth, tnheight, tcColor, tcPointStyle,    tnX1, tnY1 )

function drawLabel( tcText, tnWidth, tnHeight, tnX1, tnY1 )

enddefine


*=============================================
define class VMLArtist as AbsArtist
*=============================================

*---------------------------------------------
function transformY( ty )
return (this.nMaxY - ty)*this.nyMesh

*---------------------------------------------
function transformX( tx )
return tx*this.nXmesh


*---------------------------------------------
function drawLine( tnWidth, tcColor, tcDashStyle, tnX1, tnY1, tnX2, tnY2 )
local lcStr

lcStr = [<v:line]
lcStr = lcStr + [ strokeweight='] + alltrim(str( tnWidth,5,1)) +[px']
lcStr = lcStr + [ strokecolor='] + tcColor + [']
lcStr = lcStr + [ from='] + this.coOrdToString( this.transformX( tnX1 ), this.transformY( tnY1 ) ) + [']
lcStr = lcStr + [ to=']   + this.coOrdToString( this.transformX( tnX2 ), this.transformY( tnY2 ) ) + [']
lcStr = lcStr + [>]

if !empty(tcDashStyle) 
  lcStr = lcStr + [<v:stroke dashstyle="]+tcDashStyle+[" />]
endif

lcStr = lcStr + [</v:line>]+chr(13)+chr(10)


return lcStr

*---------------------------------------------
function drawPoint( tnWidth, tnheight, tcColor, tcPointStyle,    tnX1, tnY1 )
local lcStr, lnWidth, lnHeight

*? implement tcPointStyle
lnWidth  = tnWidth*this.nXMesh/this.nXPPU
lnHeight = tnheight*this.nYMesh/this.nYPPU

lcStr = [<v:oval style='left: ]+alltrim(str(this.transformX( tnX1)-lnwidth/2 ))+;
        [; top: ]+ alltrim(str(this.transformY( tnY1 )-lnheight/2)) + ;
        [; width:]+alltrim(str(lnWidth,10))+[; height:]+alltrim(str(lnHeight,10))+[' fillcolor="blue"/>]+chr(13)+chr(10)

return lcStr

*---------------------------------------------
function drawLabel( tcText, tcStyle, tnWidth, tnHeight, tnX1, tnY1 )
local lcStr, lnWidth, lnHeight
* width and height are in px here (not co-ords

lnWidth  = tnWidth*this.nXMesh/this.nXPPU
lnHeight = tnheight*this.nYMesh/this.nYPPU

lcStr = [<v:shape style='text-align: center; top: ]+alltrim(str(this.transformY( tnY1 )-lnheight/2))+ ;
        [; left:]+alltrim(str(this.transformX( tnX1)-lnwidth/2 ))+[;  width: ]+alltrim(str(lnWidth,10))+;
        [; height: ]+alltrim(str(lnHeight,10))+['><textbox><span class="]+alltrim(tcStyle)+[">]+;
        tcText+[</span></textbox></v:shape>]

return lcStr

*---------------------------------------------
function chartWrapper( tcElements , tcID, tnHeight, tnWidth, tnMaxX, tnMaxY, tcBGColor )
local lcStr

lcStr = [<v:group ] + ;
        [id="] + alltrim( tcID ) +[" ]+ ;
        [style="position:relative; height:]+alltrim(str( tnHeight,10 ) ) + [pt; width:]+ alltrim(str( tnWidth, 10 )) + [pt; background-color:]+ tcBGcolor +[" ] + ;
        [coordsize="] + alltrim(str( tnMaxX*this.nXMesh,10 ))+[, ]+alltrim(str( tnMaxY*this.nYMesh,10 ) )+[" coordorigin="0, 0">] + chr(13)+chr(10) + ;
        tcElements + ;
        [</v:group>]
        
return lcStr
 

enddefine 


*=============================================
define class SMGrapher as Relation
*=============================================

cID        = [SMChart]
nMaxXCoord = 100
nMaxYCoord = 100
nheight    = 100
nWidth     = 100
cBGcolor   = "white"
cAxes      = [XY]			&& determines which axes to draw
cGridLines = [XY]			&& determines which gridlines to draw
nXScale    = 10				&& interval for tics and grids
nYscale    = 10				&& interval for tics and grids
cLables    = [XY]
nYLabelScale=10
nXLabelScale=10
cTitle      =[]

dimension aDataSets(1)
oArtist = null

*---------------------------------------------
function init( tvArtist )

do case 

case vartype( tvArtist ) = "C"
  this.oArtist = createobject( tvArtist ) 
case vartype( tvAtist ) ="O"
  this.oArtist = tvArtist

endcase

*---------------------------------------------
function configureArtist()

this.oArtist.nMaxX = this.nMaxXCoord
this.oArtist.nMaxY = this.nMaxyCoord
this.oArtist.nWidth = this.nWidth
this.oArtist.nHeight = this.nHeight

*---------------------------------------------
function addDataSet( toSet )
local lnrow
 
if vartype(this.aDataSets(1)) = "L"
  lnRow = 1
else
  lnRow = alen( this.aDataSets, 1 ) + 1
  dimension this.aDataSets( lnRow )
endif

this.aDataSets( lnRow ) = toSet

*---------------------------------------------
function totalDataSets()
return iif( vartype(this.aDataSets(1))="L", 0, alen(this.aDataSets,1 ) )

*---------------------------------------------
function drawChart()
local lcStr 

this.configureArtist()

lcStr = this.drawAxes()

lcStr = lcStr + this.drawGrid()

lcStr = lcStr + this.drawDataSets()

lcstr = lcStr + this.drawTitle()

*-- wrap the vml
lcStr = this.oArtist.chartWrapper( lcStr, this.cID, this.nHeight, this.nWidth, this.nMaxXCoord, ;
                                   this.nMaxYCoord, this.cBgColor  )

return lcStr

*---------------------------------------------
function drawAxes()
local lcStr

lcStr = []

if "X" $ this.cAxes
  lcStr = lcStr + this.oArtist.drawLine(1,"black","solid", 0, 0, this.nMaxXCoord , 0) 
endif
  
if "Y" $ this.cAxes
  lcStr = lcStr + this.oArtist.drawLine(1,"black","solid", 0,0,0,this.nMaxYCoord) 
endif

return lcSTR


*---------------------------------------------
function drawLabels()
local lcStr

lcStr = []

if "X" $ this.cAxes
  lcStr = lcStr + this.oArtist.drawLine(1,"black","", 0, 0, this.nMaxXCoord , 0) 
endif
  
if "Y" $ this.cAxes
  lcStr = lcStr + this.oArtist.drawLine(1,"black",'', 0,0,0,this.nMaxYCoord) 
endif

return lcSTR

*---------------------------------------------
function drawGrid()
local lcStr, lcLabel

lcStr = []

if "X" $ this.cGridLines
  for lnk= 0 to this.nMaxXCoord step this.nXScale
    lcStr = lcStr + this.oArtist.drawLine(1,"gray","",lnk,0,lnK,this.nMaxYcoord)

  *? this is not quite right, note hard-coded significance
  if "X" $ this.cLables   and mod( round(lnK*10000,0), round(this.nXLabelScale*10000,0)) = 0
     lcStr = lcStr + this.oArtist.drawLabel( alltrim(str(lnK,10,0)), "xLabel", 40, 20, lnK, -this.nMaxYCoord*0.1 )

  endif

  endfor
endif

if "Y" $ this.cGridLines
  for lnk= 0 to this.nMaxYCoord step this.nYScale
    lcStr = lcStr + this.oArtist.drawLine(1,"gray","",0,lnK,this.nMaxXcoord, lnK)

    *? also, not scale invariant.
    *? the round, *1000 garbage is a kluge around VFP modulo bug.
    if "Y" $ this.cLables and mod( round(lnK*10000,0), round(this.nYLabelScale*10000,0)) = 0
     lcLabel =  alltrim(str(lnK,10,2))
     lcStr = lcStr + this.oArtist.drawLabel( lcLabel, "yLabel", 40, 20, -len(lcLabel)*this.nMaxXCoord*0.02, lnK-this.nMaxYCoord*0.03 )
    endif
  
  endfor
endif

return lcStr

*---------------------------------------------
function drawTitle()
local lcStr

lcStr = this.oArtist.drawLabel( This.cTitle, "charttitle", this.nwidth, this.nheight*0.15, this.nMaxXCoord/2,this.nMaxYCoord*1.10  )

return lcStr


*---------------------------------------------
function drawDataSets()
local loSet

if this.totalDataSets() = 0
  return ''
endif

local lcSTR
lcStr = []

for each loSet in this.aDataSets
  lcStr = lcStr + this.renderDataSet( loSet )
endfor 

return lcStr

*---------------------------------------------
function renderDataSet( toSet )
local lnTotal, lnK, lcStr

lnTotal = toSet.totalPoints()
if lnTotal = 0
  return ""
endif

lcStr = []
 
do case 
case toSet.cType=[XY]
  for lnK = 1 to lnTotal-1
    loPoint1 = toSet.getPoint(lnK)
    loPoint2 = toSet.getPoint(lnK+1)
    lcStr = lcStr + this.oArtist.drawLine( toSet.nLineWidth, toSet.cLineColor, toSet.cDashStyle, loPoint1.x, loPoint1.y, loPoint2.x, loPoint2.y ) 
    if !empty(toSet.cPointStyle)
      lcStr = lcStr + this.oArtist.drawPoint( 5, 5, toSet.cPointColor, toSet.cPointStyle, loPoint1.x, loPoint1.y)
    endif
  endfor  
  
  *-- draw last point
  if !empty(toSet.cPointStyle)
    lcStr = lcStr + this.oArtist.drawPoint( 5, 5, toSet.cPointColor, toSet.cPointStyle, loPoint2.x, loPoint2.y)
  endif
  
case toSet.cType=[BAR]
  for lnK = 1 to lnTotal
    loPoint1 = toSet.getPoint(lnK)
    lcStr = lcStr + this.oArtist.drawLine( toSet.nLineWidth, toSet.cLineColor,toSet.cDashStyle, loPoint1.x, 0, loPoint1.x, loPoint1.y ) 
  endfor
otherwise
  error [unrecognized dataset draw type]
endcase

return lcStr

enddefine



